#include <rtl.h>
#include <rtl_sched.h>
#include <rtl_time.h>
#include <rtl_fifo.h>
#include "rtnet.h"
#include "tcpkt.h"

//#define TCPDEBUG

#define printf rtl_printf

/* ***** Pseudo-header for UDP or TCP checksum calculation ***** */
typedef struct              /* Pseudo-header... */
{
    LWORD srce,             /* Source IP address */
          dest;             /* Destination IP address */
    BYTE  z,                /* Zero */
          pcol;             /* Protocol byte */
    WORD  len;              /* UDP length field */
} PHDR;

struct IPpass {
	int len;
	TCPKT *tcp;
};

typedef struct
{
    BYTE  vhl,
          service;          
    WORD  len;           
	struct sk_buff *skb;
} TCPKT_PATCH;


void rt_tcp_init(void); 
int rt_tcp_rcv(struct rtsocket * sk,struct sk_buff *skb);
int verify_packet(TCPKT * tcp, int len);
void swap_ip(TCPKT * tcp);
void swap_tcp(TCPKT * tcp);
WORD swapw(WORD w);
LWORD swapl(LWORD lw);
int rt_tcp_getfrag(const void *pass, char * to, unsigned int offset, unsigned int fraglen);
WORD check_tcp(TCPKT *tcp, LWORD sip, LWORD dip, int tlen);
WORD csum(void *dp, WORD count);
int print_packet(void * p);
int transmit_packet(TCPKT * tcp);
int release_packet(TCPKT * tcpkt);
int register_ootcp_fifo(int fifo_id);

static struct rtsocket *rt_tcp_socket;
int ootcp_fifo_id = -1;

int init_module(void) {

	rtl_printf("RT-Interface Loading...\n");
	rt_tcp_init();
	rtl_printf("RT-Interface Loaded.\n");

	return 0;
}


void cleanup_module(void) {
	rtl_printf("RT-Interface unloading\n");
}

int register_ootcp_fifo(int fifo_id){
	ootcp_fifo_id = fifo_id;
	return 0;
}


void rt_tcp_init(void)  {
		rt_tcp_socket = rtsocket_alloc();
		rt_tcp_socket->deliver = rt_tcp_rcv; /*********/
		rt_tcp_socket->proto = IPPROTO_TCP;  /* =6 */
		rt_inet_add_protocol(rt_tcp_socket);
}


/* H suvartnsn autn kaleitai apo to IP otav brei eva TCP paketo. */
int rt_tcp_rcv(struct rtsocket * sk,struct sk_buff *skb) {

	TCPKT * tcpkt = (TCPKT *)skb->nh.iph;
	TCPKT_PATCH * tcpkt_p = (TCPKT_PATCH *)tcpkt;

#ifdef TCPDEBUG
	rtl_printf("     ***INCOMING PACKET***\n");
#endif

	if(verify_packet(tcpkt, skb->len-20) == -1){
		rtl_printf("RT-Interface: Invalid packet received\n");
		kfree_rtskb(skb);
		return 0;
	}

	if(ootcp_fifo_id != -1){
		tcpkt_p->skb = skb;
		rtf_put(ootcp_fifo_id, &tcpkt, sizeof(TCPKT *));
	}
	else{
		rtl_printf("RT-Interface: Dropping packet - OOTCP is not currently running.\n");
		kfree_rtskb(skb);
	}

	return 0;
}


int release_packet(TCPKT * tcpkt){

	TCPKT_PATCH * tcpkt_p = (TCPKT_PATCH *)tcpkt;

	if(!tcpkt_p->skb->list)
		kfree_rtskb(tcpkt_p->skb);

	return 0;
}



void swap_tcp(TCPKT * tcp) {
	tcp->t.sport = swapw(tcp->t.sport);
	tcp->t.dport = swapw(tcp->t.dport);
	tcp->t.window = swapw(tcp->t.window);
	tcp->t.urgent = swapw(tcp->t.urgent);
	tcp->t.seq = swapl(tcp->t.seq);
	tcp->t.ack = swapl(tcp->t.ack);

}

void swap_ip(TCPKT * tcp)
{

	tcp->i.len = swapw(tcp->i.len);
	tcp->i.ident = swapw(tcp->i.ident);
	tcp->i.frags = swapw(tcp->i.frags);
	tcp->i.sip = swapl(tcp->i.sip);
	tcp->i.dip = swapl(tcp->i.dip);
}


int rt_tcp_getfrag(const void *pass, char * to, 
			unsigned int offset, unsigned int fraglen) {
	struct IPpass *p = (struct IPpass *)pass;
	TCPKT * tcp = (TCPKT *)(to - 20);
	int datalen = p->tcp->i.len - (p->tcp->t.hlen >> 2) - 20;

	if (offset != 0) {
		rtl_printf("RT-Interface: Error: IP required fragmentation.\n");
		return 0;
	}

	memcpy(to,&p->tcp->t,(p->tcp->t.hlen >> 2)); // Copy TCP header

	if(datalen > 0) // Copy data (if any)
		memcpy(&tcp->data, p->tcp->data, datalen);

	swap_tcp(tcp);
	tcp->t.check = 0;
	tcp->t.check = ~check_tcp(tcp, tcp->i.sip, tcp->i.dip, p->len - 20);

#ifdef TCPDEBUG
	print_packet(tcp);
#endif

	return 0;

}

int transmit_packet(TCPKT * tcp){
	struct IPpass p;
	struct rtsocket * sk = rt_tcp_socket;
	int len = tcp->i.len;
	int ulen = len - 20;
	struct ipcm_cookie ipc;
	struct rt_rtable * rt = NULL;
	u32 daddr;
	u32 saddr;
	int err;

	memset(&ipc,0,sizeof(ipc));
	saddr = swapl(tcp->i.sip);
	daddr = swapl(tcp->i.dip);

#ifdef TCPDEBUG
	rtl_printf("     ***OUTGOING PACKET***\n");
#endif

	err = rt_ip_route_output( &rt, daddr, saddr, 0, 0);
	
	if (err) {
		rtl_printf("RT-Interface: Error in rt_ip_route_output\n");
		return err;
	}
	p.len = len;
	p.tcp = tcp;

	err = rt_ip_build_xmit(sk,rt_tcp_getfrag,(void *)&p,
		ulen,&ipc,rt, 0);

	if (! err) {
#ifdef TCPDEBUG
		rtl_printf("RT-Interface: Packet sent to IP.\n");
#endif
		return 0;
	}
	rtl_printf("RT-Interface: Error in rt_ip_build_xmit.\n");
	return err;
}	


/* Return TCP checksum, given segment (TCP header + data) length.
 * ** The TCP segment and both IP addrs must be in network byte-order */
WORD check_tcp(TCPKT *tcp, LWORD sip, LWORD dip, int tlen)
{
    PHDR tph;
    LWORD sum;

    sum = csum(&tcp->t, (WORD)tlen);            /* Checksum TCP segment */
    tph.len = swapw((WORD)tlen);                /* Make pseudo-header */
    tph.srce = sip;
    tph.dest = dip;
    tph.z = 0;
    tph.pcol = tcp->i.pcol;
    sum += csum(&tph, sizeof(tph));             /* Checksum pseudo-header */
    return(WORD)(sum + (sum>>16));              /* Return total plus carry */
}

/* Return byte-swapped word */
WORD swapw(WORD w)
{
    return(((w<<8)&0xff00) | ((w>>8)&0x00ff));
}

/* Return byte-swapped longword */
LWORD swapl(LWORD lw)
{
    return(((lw<<24)&0xff000000L) | ((lw<<8 )&0x00ff0000L) |
            ((lw>>8 )&0x0000ff00L) | ((lw>>24)&0x000000ffL));
}

/* Do TCP-style checksum. Improved algorithm is from RFC 1071 */
WORD csum(void *dp, WORD count)
{
    register LWORD total=0L;
    register WORD n, *p, carries;
    n = count / 2;
    p = (WORD *)dp;
    while (n--)
        total += *p++;
    if (count & 1)
    	total += *(BYTE *)p;
    while ((carries=(WORD)(total>>16))!=0)
        total = (total & 0xffffL) + carries;
    return((WORD)total);
}


int print_packet(void * p){
	unsigned char *reader;
	int i;
	TCPKT * tcp = (TCPKT *)p;

//	rtl_printf("*****Printing GENFRAME*****\n");
/*
		
	rtl_printf("data: %u   ",(void *) tcp);
	rtl_printf("head: %u   ",(void *) 0);
	rtl_printf("ethh: %u\n",(void *) 0);
	rtl_printf("iphd: %u   ",(void *)tcp);
	rtl_printf("tcph: %u\n",(void *)&tcp->t);
*/

	rtl_printf("***IP***: ");
	reader = (unsigned char *) tcp;
	for (i=0;i<30;i++) {
		rtl_printf("%u  ",*reader++);
	}
	rtl_printf("\n");

	rtl_printf("***TCP***: ");
	reader = (unsigned char *) &tcp->t;
	for (i=0;i<30;i++) {
		rtl_printf("%u  ",*reader++);
	}
	rtl_printf("\n");

	return 0;
}


int verify_packet(TCPKT * tcp, int len){

	WORD sum;

	if((tcp->i.pcol != IPPROTO_TCP) || (len < sizeof(TCPHDR)) || (tcp->t.hlen < 0x50))
		return -1;

	swap_ip(tcp);

	sum=check_tcp(tcp, swapl(tcp->i.sip), swapl(tcp->i.dip), len);
	if(sum == 0xFFFF){	// if correct...
		swap_tcp(tcp);
		return 0;
	}
	else
		return -1;

}