/* Definition of TCPKT (TCP packet) */

#define BYTE __u8
#define WORD __u16
#define LWORD __u32

#define MAX_SEG_SIZE  1500	// maximum Ethernet segment size


/* ***** IP (Internet Protocol) header ***** */
typedef struct
{
    BYTE  vhl,              /* Version and header len */
          service;          /* Quality of IP service */
    WORD  len,              /* Total len of IP datagram */
          ident,            /* Identification value */
          frags;            /* Flags & fragment offset */
    BYTE  ttl,              /* Time to live */
          pcol;             /* Protocol used in data area */
    WORD  check;            /* Header checksum */
    LWORD sip,              /* IP source addr */
          dip;              /* IP dest addr */
} IPHDR;


/* ***** TCP (Transmission Control Protocol) header ***** */
typedef struct
{
    WORD  sport,            /* Source port */
          dport;            /* Destination port */
    LWORD seq,              /* Sequence number */
          ack;              /* Ack number */
    BYTE  hlen,             /* TCP header len (num of bytes << 2) */
          flags;            /* Option flags */
    WORD  window,           /* Flow control credit (num of bytes) */
          check,            /* Checksum */
          urgent;           /* Urgent data pointer */
} TCPHDR;


/* ***** TCP packet ('segment') ***** */
typedef struct
{
    IPHDR   i;                           /* IP header */
    TCPHDR  t;                           /* TCP header */
    char  *  data;						 /* Pointer to data area */
} TCPKT;
